{-# OPTIONS --type-in-type #-}

open import NeilPrelude
open import Extensionality
open import EquationalEq

-- Functors, Applicative Functors, Monads, etc...

module Class.Functor where

---------------------------------------------------------------------

record Functor (F : Set → Set) : Set where
  field fmap : {A B : Set} → (A → B) → F A → F B

        fmap-id : {{_ : Extensionality}} → {A : Set} → fmap {A} id ≡ id
        fmap-comp : {{_ : Extensionality}} → {A B C : Set} → {f : A → B} → {g : B → C} → fmap (g ∘ f) ≡ (fmap g ∘ fmap f)

---------------------------------------------------------------------

record Pointed (F : Set → Set) : Set where
  field pure : {A : Set} → A → F A

---------------------------------------------------------------------

record PointedFunctor (F : Set → Set) : Set where
  field pointedF : Pointed F
        functorF : Functor F

  open Pointed pointedF
  open Functor functorF

  field pointedLaw : {{_ : Extensionality}} → {A B : Set} → {f : A → B} → fmap f ∘ pure ≡ pure ∘ f

---------------------------------------------------------------------

record Applicative (F : Set → Set) : Set where
  field pointedfunctorF : PointedFunctor F
  open PointedFunctor pointedfunctorF using (pointedF ; functorF)
  open Pointed pointedF
  open Functor functorF

  field _<*>_ : {A B : Set} → F (A → B) → F A → F B

        fmapPureAp      : {{_ : Extensionality}} → {A B : Set} → {g : A → B} → {fa : F A} → fmap g fa ≡ pure g <*> fa
        apId            : {{_ : Extensionality}} → {A : Set} → {fa : F A} → pure id <*> fa ≡ fa
        apCompose       : {{_ : Extensionality}} → {A B C : Set} → {fa : F A} → {fbc : F (B → C)} → {fab : F (A → B)} →
                           ((pure (_∘_) <*> fbc) <*> fab) <*> fa ≡ fbc <*> (fab <*> fa)
        homomorphismLaw : {{_ : Extensionality}} → {A B : Set} → {f : A → B} → {a : A} → pure f <*> pure a ≡ pure (f a)
        interchangeLaw  : {{_ : Extensionality}} → {A B : Set} → {fab : F (A → B)} → {a : A} →
                           fab <*> pure a ≡ pure (λ g → g a) <*> fab

  infixl 4 _<*>_ _<$>_ _<∘>_

  _<$>_ : {A B : Set} → (A → B) → F A → F B
  f <$> a = pure f <*> a

  liftA2 : {A B C : Set} → (A → B → C) → F A → F B → F C
  liftA2 f a b = f <$> a <*> b

  liftA3 : {A B C D : Set} → (A → B → C → D) → F A → F B → F C → F D
  liftA3 f a b c = f <$> a <*> b <*> c

  liftA4 : {A B C D E : Set} → (A → B → C → D → E) → F A → F B → F C → F D → F E
  liftA4 f a b c d = f <$> a <*> b <*> c <*> d

  _<∘>_ : {A B C : Set} → F (B → C) → F (A → B) → F (A → C)
  _<∘>_ = liftA2 _∘_

---------------------------------------------------------------------

record Alternative (F : Set → Set) : Set where
  field applicativeF : Applicative F

        empty  : {A : Set} → F A
        _<|>_  : {A : Set} → F A → F A → F A

        empty-unitL : {{_ : Extensionality}} → {A : Set} → {fa : F A} → empty <|> fa ≡ fa
        empty-unitR : {{_ : Extensionality}} → {A : Set} → {fa : F A} → fa <|> empty ≡ fa
        mplus-assoc : {{_ : Extensionality}} → {A : Set} → {fa fb fc : F A} → fa <|> (fb <|> fc) ≡ (fa <|> fb) <|> fc

  infixl 4 _<|>_

---------------------------------------------------------------------

record Monad (M : Set → Set) : Set where
  field applicativeM : Applicative M

        join : {A : Set} → M (M A) → M A

  open Applicative applicativeM
  open PointedFunctor pointedfunctorF
  open Functor functorF
  open Pointed pointedF

  field join-assoc : {{_ : Extensionality}} → {A : Set} → join {A} ∘ join ≡ join ∘ fmap join
        join-unitL : {{_ : Extensionality}} → {A : Set} → join ∘ pure ≡ id {M A}
        join-unitR : {{_ : Extensionality}} → {A : Set} → join ∘ fmap pure ≡ id {M A}
        join-fmap  : {{_ : Extensionality}} → {A B : Set} → {f : A → B} → join ∘ fmap (fmap f) ≡ fmap f ∘ join

  infixl 4 _>>=_ _>>_ _>=>_

  _>>=_ : {A B : Set} → M A → (A → M B) → M B
  ma >>= f = join (fmap f ma)

  _>>_ : {A B : Set} → M A → M B → M B
  ma >> mb = ma >>= const mb

  _>=>_ : {A B C : Set} → (A → M B) → (B → M C) → (A → M C)
  (f >=> g) a = f a >>= g

  1stMonadLaw : {{_ : Extensionality}} → {A B : Set} → {f : A → M B} → pure >=> f ≡ f
  1stMonadLaw {A} {B} {f} = ext ( λ a →
                                      join (fmap f (pure a))
                                                             ⟨ cong join (function pointedLaw) ⟩
                                      join (pure (f a))
                                                             ⟨ function join-unitL ⟩
                                      f a
                                                             QED)

  2ndMonadLaw : {{_ : Extensionality}} → {A B : Set} → {f : A → M B} →  f >=> pure ≡ f
  2ndMonadLaw = ext (λ _ → function join-unitR)

  3rdMonadLaw : {{_ : Extensionality}} → {A B C D : Set} → {f : A → M B} → {g : B → M C} → {h : C → M D} → (f >=> g) >=> h ≡ f >=> (g >=> h)
  3rdMonadLaw {A} {B} {C} {D} {f} {g} {h} = ext (λ a →
                                                                join (fmap h (join (fmap g (f a))))
                                                                                                 ⟨ cong join (sym (function join-fmap)) ⟩
                                                                join (join (fmap (fmap h) (fmap g (f a))))
                                                                                                 ⟨ function join-assoc ⟩
                                                                join (fmap join (fmap (fmap h) (fmap g (f a))))
                                                                                                 ⟨ cong join ( fmap join (fmap (fmap h) (fmap g (f a)))
                                                                                                                                    ⟨ refl ⟩
                                                                                                               (fmap join ∘ fmap (fmap h)) (fmap g (f a))
                                                                                                                                    ⟨ sym (function fmap-comp) ⟩
                                                                                                               fmap (join ∘ fmap h) (fmap g (f a)) QED)
                                                                                                 ⟩
                                                                join (fmap (join ∘ fmap h) (fmap g (f a)))
                                                                                                 ⟨ cong join (sym (function fmap-comp)) ⟩
                                                                join (fmap (join ∘ fmap h ∘ g) (f a))
                                                                                                 QED
                                                         )

  4thMonadLaw : {{_ : Extensionality}} → {A B C : Set} → {f : A → M A} → {ma : M A} → ma >>= (pure ∘ f) ≡ fmap f ma
  4thMonadLaw {A} {B} {C} {f} {ma} = join (fmap (pure ∘ f) ma)
                                                                ⟨ cong join (function fmap-comp) ⟩
                                     join (fmap pure (fmap f ma))
                                                                ⟨ function join-unitR ⟩
                                     fmap f ma
                                                                QED

---------------------------------------------------------------------

record MonadPlus (M : Set → Set) : Set where
  field alternativeM : Alternative M
        monadM       : Monad M

  open Alternative alternativeM
  open Monad monadM

  field empty-zeroL : {{_ : Extensionality}} → {A B : Set} → {f : A → M B} → empty >>= f  ≡  empty
        empty-zeroR : {{_ : Extensionality}} → {A B : Set} → {ma : M A} → ma >> empty {B} ≡  empty

---------------------------------------------------------------------

open module FunctorInst {F} {{ functorF : Functor F }} = Functor functorF public
open module PointedInst {F} {{ pointedF : Pointed F }} = Pointed pointedF public
open module PointedFunctorInst {F} {{ pointedfunctorF : PointedFunctor F }} = PointedFunctor pointedfunctorF public
open module ApplicativeInst {F} {{ applicativeF : Applicative F }} = Applicative applicativeF public
open module AlternativeInst {F} {{ alternativeF : Alternative F }} = Alternative alternativeF public
open module MonadInst {M} {{ monadM : Monad M }} = Monad monadM public
open module MonadPlusInst {M} {{ monadplusM : MonadPlus M }} = MonadPlus monadplusM public

---------------------------------------------------------------------
